﻿#include "editor.h"
#include "ui_editor.h"
#include "globaldata.h"

Editor::Editor(QWidget *parent)
    : QMainWindow(parent),
    ui(new Ui::Editor)
{
    ui->setupUi(this);

    setupMenu();

    setupEditor();
    setupErrorLog();

    setWindowTitle(tr("BRL-CAD editor"));

    BRLCADbinPath = "/usr/brlcad/dev-7.27.0/bin/";

    this->showMaximized();
}

bool Editor::eventFilter(QObject *watched, QEvent *event)
{
    if (watched != 0 && event->type() == QEvent::Resize)
    {
        refreshRtImage();
    }
    return false;
}
void Editor::refreshRtImage()
{
    int maxImageSize = qMin(qMin(ui->lbImage->width(), ui->lbImage->height()), qMin(this->centralWidget()->height(), this->width())) - 5;
    rtImageScaled = new QPixmap(rtImage->scaled(maxImageSize, maxImageSize/*, Qt::KeepAspectRatio*/));

    ui->lbImage->setPixmap(*rtImageScaled);
    /*i++;
    ui->editorCpp->setText(QString::number(rtImageScaled->width()) + " " + QString::number(rtImageScaled->height()) + "\n"
                      + QString::number(ui->dockRaytrace->width()) + " " + QString::number(ui->dockRaytrace->height()) + "\n"
                      + QString::number(ui->lbImage->width()) + " " + QString::number(ui->lbImage->height()) + "\n" + QString::number(i)); //just for debug*/

}

//********* Setups **********

void Editor::setupMenu()
{
    QMenu *fileMenu = new QMenu(tr("&File"), this);
    menuBar()->addMenu(fileMenu);

    fileMenu->addAction(tr("&New"), this, SLOT(newFile()), QKeySequence::New);
    fileMenu->addAction(tr("&Open"), this, SLOT(openFile()), QKeySequence::Open);
    fileMenu->addAction(tr("&Save"), this, SLOT(saveFile()), QKeySequence::Save);
    fileMenu->addAction(tr("&Save As"), this, SLOT(saveFileAs()), QKeySequence::SaveAs);
    fileMenu->addAction(tr("&Export"), this, SLOT(exportAs()), QKeySequence(tr("Ctrl+E")));
    fileMenu->addAction(tr("&Import"), this, SLOT(importAs()), QKeySequence(tr("Ctrl+I")));
    fileMenu->addAction(tr("&Ratetrace"), this, SLOT(raterace()), QKeySequence(tr("F5")));

    QMenu *optionsMenu = new QMenu(tr("&Options"), this);
    menuBar()->addMenu(optionsMenu);

    optionsMenu->addAction(tr("&Properties"), this, SLOT(about()));
    optionsMenu->addAction(tr("&Shortcuts"), this, SLOT(about()));

    QMenu *windowMenu = new QMenu(tr("&Window"), this);
    menuBar()->addMenu(windowMenu);

    windowMenu->addAction(tr("Show &Logs && Errors"), ui->dockLogs, SLOT(show()));
    windowMenu->addAction(tr("Show &Properties"), ui->dockProperties, SLOT(show()));
    windowMenu->addAction(tr("Show &Raytrace"), ui->dockRaytrace, SLOT(show()));

    QMenu *helpMenu = new QMenu(tr("&Help"), this);
    menuBar()->addMenu(helpMenu);

    helpMenu->addAction(tr("&About"), this, SLOT(about()));
    helpMenu->addAction(tr("About &Qt"), qApp, SLOT(aboutQt()));
}

void Editor::setupRtImage()
{
    //ui->dockRaytrace->setWidget(ui->lbImage);

    rtImage = new QPixmap(currentPathPixmap);
    refreshRtImage();

    connect(ui->dockRaytrace, SIGNAL(allowedAreasChanged(Qt::DockWidgetAreas)), this, SLOT(refreshRtImage()));
    //connect(ui->dockRaytrace, SIGNAL(visibilityChanged(bool)), ui->dockRaytrace, SLOT(show()));

    if(CppAvailable) {
        ui->EditorsTabs->installEventFilter(this);
        ui->editorCpp->installEventFilter(this); }
    ui->editorAscii->installEventFilter(this);
    ui->lbImage->installEventFilter(this);
}

void Editor::setupEditor()
{
    QFont font;
    font.setFamily("Courier New");
    font.setFixedPitch(true);
    font.setPointSize(12);

    //ui->editorCpp = new QTextEdit();
    ui->editorCpp->setFont(font);
    ui->editorAscii->setFont(font);

    highlightercpp = new HighlighterCpp(ui->editorCpp->document());
    highlighterascii = new HighlighterAscii(ui->editorAscii->document());

   // openFile("chessBoard.cpp");
/*    QFile file("chessBoard.cpp");
    if (file.open(QFile::ReadOnly | QFile::Text))
        ui->editorCpp->setPlainText(file.readAll());*/
}


//********** File operations **********

void Editor::newFile()
{
    if(ui->editorCpp->document()->toPlainText() != "") saveFile();
    ui->editorCpp->clear();
}

void Editor::openFile(const QString &path)
{
    if(CppAvailable) { if(ui->editorCpp->document()->toPlainText() != "") saveFile(); }
    else { if(ui->editorAscii->document()->toPlainText() != "") saveFile(); }

    QString fileName = path;

    if (fileName.isNull())
        fileName = QFileDialog::getOpenFileName(this, tr("Open File"), "", "C++ Files (*.cpp *.h);;ASCII Files (*.asc);;G database (*.g)");

    if (!fileName.isEmpty())
    {
        currentPath = fileName;
        QFileInfo info(fileName);

        if (info.suffix() != "cpp") openFileAscOrG();
        else
        {
            CppAvailable = true;
            setCentralWidget(ui->EditorsTabs);

            QFile file(fileName);

            if (file.open(QFile::ReadOnly | QFile::Text))
                ui->editorCpp->setPlainText(file.readAll());
            ui->EditorsTabs->setTabText(0, file.fileName());
            file.close();
        }
    }
}

void Editor::openFileAscOrG()
{
    CppAvailable = false;
    QFileInfo info(currentPath);
    if(info.suffix() == "g") system( (BRLCADbinPath + "g2asc " + currentPath + " " + info.absoluteDir().absolutePath() + "/" + info.baseName() + ".asc").toStdString().c_str() );
    if(info.suffix() == "asc") system( (BRLCADbinPath + "asc2g " + currentPath + " " + info.absoluteDir().absolutePath() + "/" + info.baseName() + ".g").toStdString().c_str() );

    QFile ascFile(info.absoluteDir().absolutePath() + "/" + info.baseName() + ".asc");
    ascFile.open(QFile::ReadOnly | QFile::Text);
    ui->editorAscii->setPlainText(ascFile.readAll());
    ascFile.close();

    setCentralWidget(ui->editorAscii);
}

void Editor::setupErrorLog()
{
    QFont font;
    font.setFamily("Courier");
    font.setFixedPitch(true);
    font.setPointSize(10);

    ui->errorLog->setFont(font);

    //just for testing
    QFile file("CMakeOutput.log");
    if (file.open(QFile::ReadOnly | QFile::Text)) ui->errorLog->setPlainText(file.readAll());
}

void Editor::saveFile()
{
    if(currentPath.isNull()) currentPath = QFileDialog::getSaveFileName(this, tr("Open File"), "", "C++ Files (*.cpp *.h)");

    if(currentPath.isNull()) return;
    QString path = currentPath;

    /*QFile file(path);
    file.resize(0);
    fileOld.close();*/

    QFile file(path);
    file.open(QIODevice::ReadWrite);
    file.resize(0);
    QTextStream fs(&file);

    if(CppAvailable) fs << ui->editorCpp->document()->toPlainText();
    else fs << ui->editorAscii->document()->toPlainText();
    file.close();
    //openFile(currentPath);
}

void Editor::saveFileAs()
{
    QString path = QFileDialog::getOpenFileName(this, tr("Open File"), "", "C++ Files (*.cpp *.h)");

    QFile file(path);
    file.open(QIODevice::ReadWrite);
    QTextStream fs(&file);

    if(CppAvailable) fs << ui->editorCpp->document()->toPlainText();
    else fs << ui->editorAscii->document()->toPlainText();

    file.close();
}

void Editor::raterace()
{
    saveFile();
    if(currentPath.isNull()) return;
    std::cout << "rt" << std::endl << std::endl;
    //QString rtOptions = "";

    //spliting extension
    QFileInfo info(currentPath);
    QString FileName = info.absoluteDir().absolutePath() + "/" + info.baseName(); // name without extend
    QString FileNameExportedAsc = info.baseName() + ".asc"; // name without extend

    if(!CppAvailable)
    {
        system( (BRLCADbinPath + "asc2g " + FileName + ".asc " + FileName + ".g").toStdString().c_str() );
        system( (BRLCADbinPath + "rt " + ui->leOptions->text() + " -o " + FileName + ".pix " + FileName + ".g " + ui->leObjects->text()).toStdString().c_str() );
        system( ("chmod +rw " + FileName + ".pix").toStdString().c_str() ); //just for linux
        system( (BRLCADbinPath + "pix-png " + FileName + ".pix > " + FileName + ".png").toStdString().c_str() );

        currentPathPixmap = FileName + ".png";
        setupRtImage();
        return;
    }


    //compiling and running cpp
    if(CppAvailable)
    {
        if(compiler == "") compiler = "g++";
        system( (compiler + " -o " + FileName + " " + currentPath + " " + additionalFiles).toStdString().c_str() );
        system( (FileName).toStdString().c_str() );
        system(("mv " + FileNameExportedAsc + " " + FileName + ".asc").toStdString().c_str());
    }

    //convering asc to g to pix to png

    system( (BRLCADbinPath + "asc2g " + FileName + ".asc " + FileName + ".g").toStdString().c_str()); // /usr/brlcad/bin/asc2g
    system( (BRLCADbinPath + "rt " + ui->leOptions->text() + " -o " + FileName + ".pix " + FileName + ".g " + ui->leObjects->text()).toStdString().c_str() );
    system( ("chmod +rw " + FileName + ".pix").toStdString().c_str() ); //just for linux
    system( (BRLCADbinPath + "pix-png " + FileName + ".pix > " + FileName + ".png").toStdString().c_str() );

    currentPathPixmap = FileName + ".png";
    setupRtImage();

    QFile ascFile(FileName + ".asc");
    ascFile.open(QFile::ReadOnly | QFile::Text);
    ui->editorAscii->setPlainText(ascFile.readAll());
    ascFile.close();
}


//*********** Export/Import ***********

void Editor::exportAs(QString fileName)
{
    QFileInfo currentInfo(currentPath);
    QString currentFileName = currentInfo.absoluteDir().absolutePath() + "/" + currentInfo.baseName() + ".g"; // name without extend


    QString exportFileName;
    if(fileName != "") exportFileName = fileName;
    else exportFileName = QFileDialog::getSaveFileName(this, tr("Export File"), "", "asc (*.asc)"
                                                                                  ";;acad (*.acad)"
                                                                                  ";;dot (*.dot)"
                                                                                  ";;dxf (*.dxf)"
                                                                                  ";;egg (*.egg)"
                                                                                  ";;euclid (*.euclid)"
                                                                                  ";;euclid1 (*.euclid1)"
                                                                                  ";;iges (*.iges)"
                                                                                  ";;jack (*.jack)"
                                                                                  ";;nff (*.nff)"
                                                                                  ";;nmg (*.nmg)"
                                                                                  ";;obj (*.obj)"
                                                                                  ";;off (*.off)"
                                                                                  ";;ply (*.ply)"
                                                                                  ";;raw (*.raw)"
                                                                                  ";;shellrect (*.ShellRect)"
                                                                                  ";;step (*.step)"
                                                                                  ";;stl (*.stl)"
                                                                                  ";;tankill (*.tankill)"
                                                                                  ";;var (*.var)"
                                                                                  ";;voxel (*.voxel)"
                                                                                  ";;vrml (*.vrml)"
                                                                                  ";;x3d (*.x3d)");

    QFileInfo exportInfo(exportFileName);
    QString fileType = exportInfo.suffix().toLower();
    QString command;

    if(fileType == "step")
        command = BRLCADbinPath + "g-step -o " + exportFileName + " " + currentFileName;
    else if(fileType == "stl")
    {
        DialogExportSTL dialog;
        dialog.setModal(true);

        exportImportOptions = "";
        connect(&dialog, SIGNAL(optionsSet(QString)), this, SLOT(newExportImportOptions(QString)));

        dialog.exec();

        while(exportImportOptions == "");

        command = BRLCADbinPath + "g-stl " + exportImportOptions + " -o " + exportFileName + " " + currentFileName + " " + ui->leObjects->text() + " &";

        exportImportOptions = "";

    }

    std::cout << "\n\n\n" << command.toStdString() << "\n\n\n" << std::flush;

    system(command.toStdString().c_str());

    /*FILE * cmd = popen((BRLCADbinPath + exportInfo.suffix().toLower() + "-g -h").toStdString().c_str(), "r");
    QString usage;

    char buffer[255];
    while(fgets(buffer, 255, cmd)) usage += QString(buffer);
    fclose(cmd);*/

    //system(QInputDialog::getText(this, "Export Options", "", QLineEdit::Normal, (BRLCADbinPath + exportInfo.suffix().toLower() + "-g -o" + exportFileName + " " + currentFileName + ".g " + ui->leObjects->text())).toStdString().c_str());

    //system((BRLCADbinPath + "g-" + exportInfo.suffix().toLower() + " -o " + exportFileName + " " + currentFileName + ".g " + ui->leObjects->text()).toStdString().c_str() );
}

void Editor::importAs(QString fileName)
{
    QFileInfo currentInfo(currentPath);
    QString currentDir = currentInfo.absoluteDir().absolutePath() + "/"; // name without extend

    QString importFileName = QFileDialog::getOpenFileName(this, tr("Open File"), "", "asc (*.asc)"
                                                                         ";;3dm (*.3dm)"
                                                                         ";;comgeom (*.comgeom)"
                                                                         ";;vg (*.vg)"
                                                                         ";;cy (*.cy)"
                                                                         ";;dem (*.dem)"
                                                                         ";;dxf (*.dxf)"
                                                                         ";;enf (*.enf)"
                                                                         ";;euclid (*.euclid)"
                                                                         ";;fast4 (*.fast4)"
                                                                         ";;iges (*.iges)"
                                                                         ";;jack (*.jack)"
                                                                         ";;k (*.k)"
                                                                         ";;nastran (*.nastran)"
                                                                         ";;obj (*.obj)"
                                                                         ";;off (*.off)"
                                                                         ";;patch (*.patch)"
                                                                         ";;ply (*.ply)"
                                                                         ";;proe (*.proe)"
                                                                         ";;raw (*.raw)"
                                                                         ";;shp (*.shp)"
                                                                         ";;step (*.step)"
                                                                         ";;stl (*.stl)"
                                                                         ";;tankill (*.tankill)"
                                                                         ";;tgf (*.tgf)"
                                                                         ";;viewpoint (*.viewpoint)");

    QFileInfo importInfo(importFileName);
    system((BRLCADbinPath + importInfo.suffix().toLower() + "-g " + importFileName + " -o " + currentDir + "tmp.g").toStdString().c_str() );

    system((BRLCADbinPath + "g2asc " + currentDir + "tmp.g " + currentDir + "tmp.asc").toStdString().c_str() );

    QFile ascImport(currentDir + "tmp.asc");
    ascImport.open(QFile::ReadOnly);
    ui->editorAscii->setPlainText(ui->editorAscii->document()->toPlainText() + "\n" + QString::fromStdString(ascImport.readAll().toStdString()));
    ascImport.close();

}





/********** Not Used **********/

/*
 -r		Report overlaps (default)
 -R		Do not report overlaps
 -M		Read matrix+commands on stdin
 -o model.pix	Output .pix format file (default is window)
 -s #		Square grid size in pixels (default is 512)
 -w # -n #	Grid size width (w) and height (n) in pixels
 -a # -e #	Azimuth (a) and elevation (e) in degrees
 -V #		View (pixel) aspect ratio (width/height)
 -p #		Perspective angle, degrees side to side (0 <= # < 180)
 -P #		Set number of processors
 -T # or -T #,# or -T #/#
        Tolerance: distance or distance,angular or distance/angular
 -l #		Set lighting model rendering style (default is 0)
 -U #		Use air if # is greater than 0
 -x #		librt debug flags
 -N #		NMG debug flags
 -X #		rt debug flags
 -. #		Select factor in NUgrid algorithm (default is 1.5)
 -, #		Selection of which space partitioning algorithm to use
 -@ #		Set limit to each dimension of the nugrid
 -b "# #"	Specify X and Y pixel coordinates (need quotes) for single ray to be fired, for debugging
 -c script_command
        Supply, on command line, command which can appear in -M command script
 -d #		Set flag for reporting of pixel distances
 -f #		Set expected playback rate in frames-per-second (default is 30)
 -g #		Set grid cell width, in millimeters
 -m density,r,g,b
        Provide parameters for an exponential shading (default r,g,b is 0.8,0.9,0.99)
 -i		Enable incremental mode processing
 -j xmin,xmax,ymin,ymax
        Enable processing of sub-rectangle
 -k xdir,ydir,zdir,dist
        Enable use of a cutting plane
 -l #		Select lighting model (default is 0)
 -t		Reverse the order of grid traversal (default is not to do that)
 -u units	Specify the units (or use "model" for the local model's units)
 -v #		Set the verbosity bit vector flags
 -A #		Set the ambient light intensity
 -B		Turn on the "benchmark" flag (default is off)
 -C #/#/#	Set the background color to the RGB value #/#/#
 -D #		Specify the starting frame number (ending frame number is specified via -K #)
 -E #           Set the distance from eye point to center of the model RPP (default is sqrt(2))
 -F framebuffer	Cause output to be sent to the indicated framebuffer
 -G #		Set grid cell height, in millimeters
 -H #		Set number of extra rays to fire
 -J #		Set a bit vector for "jitter"
 -K #		Specify the ending frame number (starting frame number is specified via -D #)
 -O model.pix	Output .pix format file, double precision format
 -Q x,y		Select pixel ray for query with debugging; compute other pixels without debugging
 -S		Enable stereo viewing (off by default)
 -W		Set background image color to white (default is black)
 -! #		Turn on the libbu(3) library debugging flags
 -+ t		Specify that output is NOT binary (default is that it is); -+ is otherwise not
        implemented
*/


















